#include "xmlpatch.h"

XmlPatch::XmlPatch(QString patchFilesWildcard, QString forceTargetFilesWildcard, bool noBackups)
{
    this->patchFilesToProcess=UtilXmlTools::getAllPatchFilesByWildcard(patchFilesWildcard);
    this->forceTargetFilesWildcard=forceTargetFilesWildcard;
    this->backupsEnabled=!noBackups;

    if(forceTargetFilesWildcard!=""){
        std::cout << "User forced patch in the target file(s): " << this->forceTargetFilesWildcard.toLatin1().constData() << std::endl;
    }

    if(this->patchFilesToProcess.size()==0){
        UtilXmlTools::displayErrorMessage("Loading patch files","No .patch or .oni-patch files were found for the wildcard: "+patchFilesWildcard);
    }

}

void XmlPatch::readAndProcessPatchFile(){

    // Process all PatchFiles
    for(int i=0; i<this->patchFilesToProcess.size(); i++){

        QFile inputFile(this->patchFilesToProcess[i]);

        if (inputFile.open(QIODevice::ReadOnly))
        {

            QTextStream fileStream(&inputFile);

            checkPatchVersion(this->patchFilesToProcess[i], fileStream);
            checkAndProcessValidCommands(fileStream);

            inputFile.close();
        }
        else{
            UtilXmlTools::displayErrorMessage("Read patch file", "Error opening patch file: '" + this->patchFilesToProcess[i] + "'.\n" + inputFile.errorString());
        }

    }

    UtilXmlTools::displaySuccessMessage(this->patchFilesToProcess.size(),"Patch File(s)");

}

void XmlPatch::addToOperation(const QString &xmlString, XmlFilter &filters, const QString &filesWildcard){

    QStringList filesToProcess;

    filesToProcess=UtilXmlTools::getAllXmlFilesByWildcard(filesWildcard);

    pugi::xml_node nodeToInsertion;
    pugi::xml_document newNode;
    pugi::xml_parse_result result;

    result=newNode.load(xmlString.toLatin1().constData()); // load xml to insert

    if(result.status!=pugi::status_ok){
        UtilXmlTools::displayErrorMessage("@ADD_INSIDE_NODE", "The xml node to insert is invalid.\n" + Util::toQString(result.description()));
    }

    // Process all XmlFiles
    for(int i=0; i<filesToProcess.size(); i++){

        UtilXmlTools::loadXmlFile(filesToProcess[i],this->document,this->rootNode,this->backupsEnabled,"@ADD_INSIDE_NODE");

        nodeToInsertion=UtilXmlTools::getFirstNamedElements(this->rootNode,filters);

        if(nodeToInsertion.type()==pugi::node_null){

            QString errMessage;

            errMessage = "It wasn't found a node with a ElementName: '" + filters.getElementName() + "'";
            if(filters.getParentElementName()!=""){
                errMessage += " and a ParentElementName: '" + filters.getParentElementName() + "'";
            }
            if(filters.getAttributeName()!=""){
                errMessage += " and an AttributeName: '" + filters.getAttributeName() + "' and an AttributeValue: '" + filters.getAttributeValue() + "'";
            }

            UtilXmlTools::displayErrorMessage("@ADD_INSIDE_NODE",errMessage);
        }

        nodeToInsertion.prepend_copy(newNode.first_child()); // append the new node


        UtilXmlTools::saveXmlFile(filesToProcess[i],this->document, "@ADD_INSIDE_NODE");
    }

    UtilXmlTools::displaySuccessMessage(filesToProcess.size(),"@ADD_INSIDE_NODE");
}

void XmlPatch::removeNodeOperation(XmlFilter &filters, const QString &filesWildcard){

    QStringList filesToProcess;

    filesToProcess=UtilXmlTools::getAllXmlFilesByWildcard(filesWildcard);

    pugi::xml_node nodeToDeletion;
    pugi::xml_parse_result result;

    // Process all XmlFiles
    for(int i=0; i<filesToProcess.size(); i++){

        UtilXmlTools::loadXmlFile(filesToProcess[i],this->document,this->rootNode,this->backupsEnabled,"@REMOVE_NODE");

        nodeToDeletion=UtilXmlTools::getFirstNamedElements(this->rootNode,filters);

        if(nodeToDeletion.type()==pugi::node_null){
            QString errMessage;

            errMessage = "It wasn't found a node with a ElementName: '" + filters.getElementName() + "'";
            if(filters.getParentElementName()!=""){
                errMessage += " and a ParentElementName: '" + filters.getParentElementName() + "'";
            }
            if(filters.getAttributeName()!=""){
                errMessage += " and an AttributeName: '" + filters.getAttributeName() + "' and an AttributeValue: '" + filters.getAttributeValue() + "'";
            }

            UtilXmlTools::displayErrorMessage("@REMOVE_NODE",errMessage);
        }

        if(!nodeToDeletion.parent().remove_child(nodeToDeletion)){  // remove the node

            QString errMessage;

            errMessage = "Couldn't remove the node with Element '" + filters.getElementName() + "'";

            if(filters.getParentElementName()!=""){
                errMessage += " and a ParentElement: '" + filters.getParentElementName() + "'";
            }

            UtilXmlTools::displayErrorMessage("@REMOVE_NODE",errMessage);
        }

        UtilXmlTools::saveXmlFile(filesToProcess[i],this->document, "@REMOVE_NODE");
    }

    UtilXmlTools::displaySuccessMessage(filesToProcess.size(), "@REMOVE_NODE");
}

void XmlPatch::executeCommandOperation(const QString &commandString){

    QProcess newXmlToolsInstance;
    QString resultOutput;

    // Avoid infinite fork loops
    if(commandString.contains("-p") || commandString.contains("--patch-files")){
        UtilXmlTools::displayErrorMessage("@COMMAND","Use of --patch-files option is not allowed inside a patch file");
    }

    newXmlToolsInstance.start(GlobalVars::AppName + " " + commandString);
    newXmlToolsInstance.waitForFinished(-1); // wait for new instance to finish

    resultOutput=newXmlToolsInstance.readAll();

    if(newXmlToolsInstance.exitCode()!=0){
        UtilXmlTools::displayErrorMessage("@COMMAND", "An error ocurred:\n" + resultOutput);
        exit(1);
    }

    std::cout << "@COMMAND patch operation output:\n" << resultOutput.toLatin1().constData() << std::endl;

    UtilXmlTools::displaySuccessMessage(1,"@COMMAND");
}

void XmlPatch::executeCustomCommandOperation(const QString &jsString, const QString &filesWildcard){

    QStringList filesToProcess;
#ifdef _USE_OLD_JS_ENGINE
    QScriptEngine engine;
    QScriptValue engineResult; // variable to check for js_errors
#else
    QJSEngine engine;
    QJSValue engineResult; // variable to check for js_errors
#endif

    QString rexmlString, jsxmlString, currXmlFileString;


    QFile rexmlfile(":/resources/libs/rexml.js");
    QFile jsxmlfile(":/resources/libs/jsxml.js");

    filesToProcess=UtilXmlTools::getAllXmlFilesByWildcard(filesWildcard);

    rexmlfile.open(QFile::ReadOnly | QFile::Text);
    jsxmlfile.open(QFile::ReadOnly | QFile::Text);

    rexmlString=QTextStream(&rexmlfile).readAll();
    jsxmlString=QTextStream(&jsxmlfile).readAll();

    engine.evaluate(rexmlString); // load js libraries
    engine.evaluate(jsxmlString);

    // Process all XmlFiles
    for(int i=0; i<filesToProcess.size(); i++){

        if(this->backupsEnabled){
            UtilXmlTools::backupFile(filesToProcess[i]);
        }

        QFile currXmlFile(filesToProcess[i]);

        if(!currXmlFile.open(QFile::ReadOnly | QFile::Text)){
            UtilXmlTools::displayErrorMessage("@CUSTOM_CODE","Error loading '" + filesToProcess[i] + "' file for read operation.");
        }

        currXmlFileString=QTextStream(&currXmlFile).readAll();

        currXmlFile.close(); // close reading

        engine.globalObject().setProperty("$xmlData",currXmlFileString);

        engineResult=engine.evaluate(jsString);

#ifdef _USE_OLD_JS_ENGINE
        if (engine.hasUncaughtException()) {
            displayJsException(engine,engineResult);
        }
#else
        checkForJsException(engineResult);
#endif

        if(!currXmlFile.open(QFile::WriteOnly | QFile::Text | QIODevice::Truncate)){
            UtilXmlTools::displayErrorMessage("@CUSTOM_CODE","Error loading '" + filesToProcess[i] + "' file for @CUSTOM_CODE write operation.");
        }

        engineResult=engine.globalObject().property("$xmlData");

#ifdef _USE_OLD_JS_ENGINE
        if (engine.hasUncaughtException()) {
            displayJsException(engine,engineResult);
        }
#else
        checkForJsException(engineResult);
#endif

        QTextStream(&currXmlFile) << engineResult.toString(); // retreive the modified xml by javascript and save it to the file
    }

    UtilXmlTools::displaySuccessMessage(filesToProcess.size(), "@CUSTOM_CODE");
}

void XmlPatch::checkPatchVersion(const QString &file, QTextStream &fileStream){

    QString line, patchVersion="";

    // First get the patch version and check it validity
    while ( !fileStream.atEnd() ){
        line = fileStream.readLine();

        if(line.startsWith('#')){ // Ignore comments
            continue;
        }
        else if(line.startsWith("@XML_TOOLS")){

            patchVersion=getPatchParameterValue(line,"Version");

            if(!patchVersion.startsWith("2.0")){
                QString errMessage;

                errMessage = "The current patch version is incompatible with this XmlTools version:\n";
                errMessage += "Patch file name: '" + file + "'\n";
                errMessage += "XmlTools version:  " + GlobalVars::AppVersion + "\n" + "CurrPatch version: " + patchVersion + "";
                UtilXmlTools::displayErrorMessage("@XML_TOOLS",errMessage);
            }
            break; // We have got what we wanted
        }
    }

    if(patchVersion==""){
        UtilXmlTools::displayErrorMessage("@XML_TOOLS","Patch version not found.");
    }

}

void XmlPatch::checkAndProcessValidCommands(QTextStream &fileStream){

    QString line;
    QString filesWildcard;
    QString xmlToInsert, command, jsCode;
    XmlFilter filters;

    // Process the rest of the commands in patch file
    while ( !fileStream.atEnd() ){
        line = fileStream.readLine();

        if(line.startsWith('#')){ // Ignore comments
            continue;
        }
        else if(line.startsWith("@ADD_INSIDE_NODE")){
            filters.setElementName(getPatchParameterValue(line,"ElementName"));
            filters.setParentElementName(getPatchParameterValue(line,"ParentElementName"));
            filters.setAttributeName(getPatchParameterValue(line,"AttributeName"));
            filters.setAttributeValue(getPatchParameterValue(line,"AttributeValue"));
            if(this->forceTargetFilesWildcard==""){
                filesWildcard=getPatchParameterValue(line,"Files");
            }
            else{
                filesWildcard=this->forceTargetFilesWildcard;
            }

            // Check attribute filters
            if(filters.getAttributeName()!="" && filters.getAttributeValue()==""){
                UtilXmlTools::displayErrorMessage("@ADD_INSIDE_NODE","attribute-value option is required if using attribute-name option.");
            }

            if(filters.getAttributeValue()!="" && filters.getAttributeName()==""){
                UtilXmlTools::displayErrorMessage("@ADD_INSIDE_NODE","attribute-name option is required if using attribute-value option.");
            }

            while ( !fileStream.atEnd() && !line.startsWith("</xml>")){
                line = fileStream.readLine();

                if(!line.startsWith("<xml>") && !line.startsWith("</xml>")){
                    xmlToInsert += line + "\n";
                }
            }

            addToOperation(xmlToInsert,filters,filesWildcard);

        }
        else if(line.startsWith("@REMOVE_NODE")){

            filters.setElementName(getPatchParameterValue(line,"ElementName"));
            filters.setParentElementName(getPatchParameterValue(line,"ParentElementName"));
            filters.setAttributeName(getPatchParameterValue(line,"AttributeName"));
            filters.setAttributeValue(getPatchParameterValue(line,"AttributeValue"));
            if(this->forceTargetFilesWildcard==""){
                filesWildcard=getPatchParameterValue(line,"Files");
            }
            else{
                filesWildcard=this->forceTargetFilesWildcard;
            }

            // Check attribute filters
            if(filters.getAttributeName()!="" && filters.getAttributeValue()==""){
                UtilXmlTools::displayErrorMessage("@REMOVE_NODE","attribute-value option is required if using attribute-name option.");
            }

            if(filters.getAttributeValue()!="" && filters.getAttributeName()==""){
                UtilXmlTools::displayErrorMessage("@REMOVE_NODE","attribute-name option is required if using attribute-value option.");
            }

            removeNodeOperation(filters,filesWildcard);

        }
        else if(line.startsWith("@COMMAND")){

            command=getPatchParameterValue(line,"");

            executeCommandOperation(command);
        }
        else if(line.startsWith("@CUSTOM_CODE")){

            if(this->forceTargetFilesWildcard==""){
                filesWildcard=getPatchParameterValue(line,"Files");
            }
            else{
                filesWildcard=this->forceTargetFilesWildcard;
            }

            while ( !fileStream.atEnd() && !line.startsWith("</code>")){

                line = fileStream.readLine();

                if(!line.startsWith("<code>") && !line.startsWith("</code>")){
                    jsCode += line + "\n";
                }
            }

            executeCustomCommandOperation(jsCode,filesWildcard);
        }
    }
}

QString XmlPatch::getPatchParameterValue(const QString& line, QString parameter){

    int startValueIdx, endValueIdx;

    parameter+=" "; // Parameters include a space before it's value.

    if(!line.contains(parameter)){
        if(parameter=="ParentElementName " || parameter=="AttributeName " || parameter=="AttributeValue "){
            return ""; // not mandatory
        }
        parameter.remove(" "); // just remove the space added so it doesn't look weird when outputted to the user

        UtilXmlTools::displayErrorMessage("Read patch file parameter","Couldn't retreive '" + parameter + "' parameter.");
    }

    startValueIdx=line.indexOf(parameter); // get the position where parameter is defined
    endValueIdx=line.indexOf("\"",startValueIdx)+1; // get the position where parameter value begins (+1 to ignore mandotory quote)

    startValueIdx=endValueIdx;
    endValueIdx=line.indexOf("\"",startValueIdx); // get the last mandatory quote of the value

    return line.mid(startValueIdx,endValueIdx-startValueIdx); // return the value of the parameter (is in the middle of the mandatory quotes)
}

#ifdef _USE_OLD_JS_ENGINE
void XmlPatch::displayJsException(QScriptEngine &engine, QScriptValue &engineResult){
    if (engine.hasUncaughtException()) {
        UtilXmlTools::displayErrorMessage("@CUSTOM_CODE","Uncaught js exception (user code) at line " +QString::number(engine.uncaughtExceptionLineNumber()) + ":\n" + engineResult.toString());
    }
}
#else
void XmlPatch::checkForJsException(QJSValue &engineResult){
    if (engineResult.isError()) {
        UtilXmlTools::displayErrorMessage("@CUSTOM_CODE","Uncaught js exception (user code):\n" + engineResult.toString());
    }
}
#endif
